# -*- coding: UTF-8 -*-
#
# Здесь происходит вся работа с очередью задач на выгрузку
#


from artix.common import py2to3utils
from plugins import backupFile, status
from propertyReader import properties
from snakemq.message import Message, FLAG_PERSISTENT
from snakemq.queues import QueuesManager
from snakemq.storage.sqlite import SqliteQueuesStorage
import json
import logger
import os
import sendSalesModule as sendSalesModule
import threading
import traceback

loggerSync = logger.getLogger()

class Synchronizer:
    def __init__(self, lock, type):
        self.lock = lock

    def __enter__(self):
        self.lock.acquire()
        return self.lock

    def __exit__(self, type, value, traceback):
        self.lock.release()

__TASK_HANDLE_SHORT_SLEEP_TIME = 3
__TASK_HANDLE_LONG_SLEEP_TIME = 60

unloadTasksLocker = Synchronizer(threading.Lock(), "unloadTasksLocker")
unloadedSalesTasksLocker = Synchronizer(threading.Lock(), "unloadedSalesTasksLocker")

unloadTasksQueueName = 'unloadTasks'  # имя очереди задач на выгрузку продаж (в ней обрабатываются таски на выгрузку продаж)
unloadedSalesTasksQueueName = 'unloadedSalesTasks'  # имя очереди задач на выгрузку сгенерированных продаж (архив с продажами сформирован, но еще не передан), и помещены в особую директорию

THROUGH_GRPC = 0    # отправка продаж будет происходить через grpc
THROUGH_QPID = 1    # отправка продаж будет происходить через Qpid


def getQueueByName(name):
    """ Получение очереди задач по имени """
    return QueuesManager(SqliteQueuesStorage(name + '.db')).get_queue(name)


def checkUnloadSalesOnSkipMode():
    """
    Если очередь недооправленных продаж оставлена по ошибке, то в каждый объект записывается поле skip=True (не спрашивайте почему в каждое сообщение,
    видимо так повелось), и в данном случае попытка отправки приостанавливается до получения "внешнего" события (перезапуск sync или запроса с КС через thrift)
    Данная функция проверяет взведение этого флага у первого объекта в очереди, и если он там есть, возвращает status.skipped, иначе возвращает None
    :return: возвращает status.skipped, если он в первой записи == True, иначе возвращает None
    """
    returnStatus = None
    with unloadedSalesTasksLocker:
        queueTasks = getQueueByName(unloadedSalesTasksQueueName)
        if len(queueTasks) > 0:
            task = json.loads(py2to3utils.ensureStr(queueTasks.get().data))
            if task.get('skip'):
                returnStatus = status.skipped
        queueTasks.manager.close()
    return returnStatus


def checkUnloadSalesQueueNotEmpty():
    """
    Проверка, что в очереде есть задачи на выгрузку продаж
    :return: возвращает True, если в очереде есть задачи на выгрузку продаж, иначе возвращает False
    """
    with unloadTasksLocker:
        queueTasks = getQueueByName(unloadTasksQueueName)
        return True if len(queueTasks) > 0 else False


def handleTasks(mode):
    """
    Обработка очереди продаж на выгрузку
    :param mode - способ отправки продаж (через GRPC / через QPID)
    """
    nextSleep = __TASK_HANDLE_SHORT_SLEEP_TIME
    res = status.empty
    with unloadTasksLocker:
        queueTasks = getQueueByName(unloadTasksQueueName)
        for taskIndex in range(len(queueTasks)):
            taskStr = queueTasks.get().data
            if taskStr is not None:
                task = json.loads(py2to3utils.ensureStr(taskStr))

                res = sendSalesModule.unloadRun(task)
                if res == status.busy:
                    nextSleep = __TASK_HANDLE_SHORT_SLEEP_TIME
                    break
                elif mode == THROUGH_QPID and res in [status.errorQpidInsert, status.errorUnload]:
                    queueTasks.pop()
                    task.update({"skip": True})
                    with unloadedSalesTasksLocker:
                        unloadedSalesTasksQueue = getQueueByName(unloadedSalesTasksQueueName)
                        unloadedSalesTasksQueue.push(Message(json.dumps(task).encode('utf-8'), ttl=60, flags=FLAG_PERSISTENT))
                        unloadedSalesTasksQueue.manager.close()
                    continue
                elif mode == THROUGH_QPID and res == status.OK:
                    queueTasks.pop()
                    continue
                elif mode == THROUGH_GRPC and res in [status.start, status.OK]:
                    queueTasks.pop()
                    task.update({"skip": False})
                    with unloadedSalesTasksLocker:
                        unloadedSalesTasksQueue = getQueueByName(unloadedSalesTasksQueueName)
                        unloadedSalesTasksQueue.push(Message(json.dumps(task).encode('utf-8'), ttl=60, flags=FLAG_PERSISTENT))
                        unloadedSalesTasksQueue.manager.close()
                    continue
                else:
                    nextSleep = __TASK_HANDLE_LONG_SLEEP_TIME
                    break
        queueTasks.manager.close()
    return nextSleep, res


def removeUnloadedSalesByGrpc(successUnloadedSalesIds):
    """
    Функция исключения завершенных тасков и очереди выгрузки продаж. В качестве параметра передаются идентификаторы продажи, которые были отправлены на КС.
    Если КС их успешно обработал (по его мнению конечно же), то этот список должен содержать идентификтор этого таска
    Соответсвенно, если идентификатор задачи найден в списке - она удаляется.
    :param removeAll - удалить все таски, т.к. отложенных задач не было найдено в директории
    """
    with unloadedSalesTasksLocker:
        # можно удалить из очереди выгруженных продаж все записи, если директория с ними пуста
        removeAll = os.listdir(backupFile.getUnloadedSalesPath()) == []
        if successUnloadedSalesIds or removeAll:
            loggerSync.debug('Success task ids: %s', successUnloadedSalesIds)
            unloadedSalesTasksQueue = getQueueByName(unloadedSalesTasksQueueName)
            for taskIndex in range(len(unloadedSalesTasksQueue)):
                task = json.loads(py2to3utils.ensureStr(unloadedSalesTasksQueue.get().data))
                if removeAll:
                    unloadedSalesTasksQueue.pop()
                    loggerSync.debug('Remove task with Id: %s from task queue', task.get('taskId'))
                else:
                    if task.get('taskId') in successUnloadedSalesIds:
                        loggerSync.debug('Remove task with Id: %s from task queue', task.get('taskId'))
                        unloadedSalesTasksQueue.pop()
            loggerSync.debug('Removing tasks operation success')
            unloadedSalesTasksQueue.manager.close()


def checkUnloadedSalesQpid():
    """
    Функция проверки очереди невыгруженных продаж.
    1. Если в очереди есть задачи и её поле 'skip' равно True, то возвращаем status.skipped
    2. Если в очереди есть задачи и её поле 'skip' равно False или отсутствует, то выполняем выгрузку невыгруженной продажи для этой задачи.
       В случае успешной выгрузки задача удаляется из очереди
    """
    queueTasks = None
    unloadResult = status.OK
    with unloadedSalesTasksLocker:
        try:
            queueTasks = getQueueByName(unloadedSalesTasksQueueName)
            for taskIndex in range(len(queueTasks)):
                task = json.loads(py2to3utils.ensureStr(queueTasks.get().data))
                unloadResult = unloadedSalesRun(task)
                if unloadResult != status.OK:
                    break
                queueTasks.pop()
        except Exception as e:
            loggerSync.error("{0}\n{1}".format(str(e), traceback.format_exc()))
        finally:
            if queueTasks is not None and queueTasks.manager:
                queueTasks.manager.close()
    return unloadResult


def unloadedSalesRun(task):
    """ Запуск выгрузки невыгруженных продаж """
    # reconfigure(task.get('configInfo')) # а нужен ли?
    return sendSalesModule.unloadedSales(properties.localBroker, properties.mode, task)


def updateUnloadedSalesTaskState(skip):
    """
    Функция устанавливает значение поля 'skip' для всех задач очереди невыгруженных продаж равным значению передаваемого параметра skip
    :return список "плохих" тасков
    """
    tasksList = []
    try:
        with unloadedSalesTasksLocker:
            queueTasks = getQueueByName(unloadedSalesTasksQueueName)
            for taskIndex in range(len(queueTasks)):
                taskStr = queueTasks.get().data
                if taskStr is not None:
                    task = json.loads(py2to3utils.ensureStr(taskStr))
                    task.update({'skip': skip})
                    queueTasks.pop()
                    queueTasks.push(Message(json.dumps(task).encode('utf-8'), ttl=60, flags=FLAG_PERSISTENT))
                    tasksList.append(task)
            queueTasks.manager.close()
        if skip is True:
            loggerSync.info("Выгрузка продаж остановлена из-за ошибки в передаче данных, повторная попытка произойдет после перезапуска sync-а"
                            " или запроса на выгрузку продаж с КС")
    except Exception as e:
        loggerSync.error("{0}\n{1}".format(str(e), traceback.format_exc()))
    return tasksList


def getUnloadTasksQueueSize():
    """
    Получить размер очереди unloadTasks.db
    """
    count = None
    with unloadTasksLocker:
        queue = getQueueByName(unloadTasksQueueName)
        count = len(queue)
        queue.manager.close()
    return count


def getUnloadedSalesTasksQueueSize():
    """
    Получить размер очереди unloadedSalesTasks.db
    """
    count = None
    with unloadedSalesTasksLocker:
        queue = getQueueByName(unloadedSalesTasksQueueName)
        count = len(queue)
        queue.manager.close()
    return count
